#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Sans IO protocol handling code to the GA4GH streaming API.
"""
from __future__ import division
from __future__ import print_function

import base64
import logging
import time

from six.moves.urllib.parse import urlencode
from six.moves.urllib.parse import urlunparse
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import parse_qs

import htsget.exceptions as exceptions


def ticket_request_url(
        url, fmt=None, reference_name=None, reference_md5=None,
        start=None, end=None, fields=None, tags=None, notags=None,
        data_format=None):
    parsed_url = urlparse(url)
    get_vars = parse_qs(parsed_url.query)
    # TODO error checking
    if reference_name is not None:
        get_vars["referenceName"] = reference_name
    if reference_md5 is not None:
        get_vars["referenceMd5"] = reference_md5
    if start is not None:
        get_vars["start"] = int(start)
    if end is not None:
        get_vars["end"] = int(end)
    if data_format is not None:
        get_vars["format"] = data_format.upper()
    # if fields is not None:
    #     get_vars["fields"] = ",".join(fields)
    # if tags is not None:
    #     get_vars["tags"] = ",".join(tags)
    # if notags is not None:
    #     get_vars["notags"] = ",".join(notags)
    new_url = list(parsed_url)
    new_url[4] = urlencode(get_vars, doseq=True)
    return urlunparse(new_url)


class DownloadManager(object):
    """
    Abstract implementation of the protocol.
    """
    def __init__(
            self, url, output, data_format=None, reference_name=None,
            reference_md5=None, start=None, end=None, fields=None, tags=None,
            notags=None, max_retries=5, timeout=10, retry_wait=5):
        self.max_retries = max_retries
        self.timeout = timeout
        self.retry_wait = retry_wait
        self.output = output
        self.ticket_request_url = ticket_request_url(
            url, data_format=data_format, reference_name=reference_name,
            reference_md5=reference_md5, start=start, end=end, fields=fields,
            tags=tags, notags=notags)
        self.ticket = None
        self.data_format = format
        self.md5 = None

    def __retry(self, method, *args):
        completed = False
        num_retries = 0
        position_before = None
        try:
            # stdout does not support seek/tell, so we disable retry if this fails
            position_before = self.output.tell()
        except IOError:
            pass
        while not completed:
            try:
                method(*args)
                completed = True
            except exceptions.RetryableError as re:
                if position_before is not None and num_retries < self.max_retries:
                    num_retries += 1
                    sleep_time = self.retry_wait  # TODO exponential backoff
                    logging.warning(
                        "Error: '{}' occured; sleeping {}s before retrying "
                        "(attempt={})".format(re, sleep_time, num_retries))
                    self.output.seek(position_before)
                    time.sleep(sleep_time)
                else:
                    raise re

    def _ticket_request(self):
        raise NotImplementedError()

    def _handle_data_uri(self, parsed_url):
        split = parsed_url.path.split(",", 1)
        # TODO parse out the encoding properly.
        description = split[0]
        data = base64.b64decode(split[1])
        logging.debug("handle_data_uri({}, length={})".format(description, len(data)))
        self.output.write(data)

    def _handle_http_url(url, headers):
        raise NotImplementedError()

    def run(self):
        self.__retry(self._handle_ticket_request)
        self.data_format = self.ticket.get("format", "BAM")
        self.md5 = self.ticket.get("md5", None)
        for url_object in self.ticket["urls"]:
            url = urlparse(url_object["url"])
            if url.scheme.startswith("http"):
                headers = url_object.get("headers", "")
                self.__retry(self._handle_http_url, urlunparse(url), headers)
            elif url.scheme == "data":
                self._handle_data_uri(url)
            else:
                raise ValueError("Unsupported URL scheme:{}".format(url.scheme))



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Exceptions for htsget.
"""
from __future__ import division
from __future__ import print_function


class HtsgetException(Exception):
    """
    Superclass of all exceptions that Htsget raises directly.
    """


class ExceptionWrapper(HtsgetException):
    """
    A wrapper for exceptions raised by lower-level libraries. The source
    exception is stored in the ``source`` instance variable.
    """
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return str(self.source)


class RetryableError(HtsgetException):
    """
    The superclass of all errors that we think are worth retrying.
    """


class RetryableIOError(RetryableError, ExceptionWrapper):
    """
    All exceptions thrown by lower level libraries that we consider to be
    retryable are wrapped by this exception.
    """


class ContentLengthMismatch(RetryableError):
    """
    The length of the downloaded content is not the same as the
    length reported in the header.
    """



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import division
from __future__ import print_function

__version__ = "undefined"
try:
    from . import _version
    __version__ = _version.version
except ImportError:
    pass

from .io import get  # NOQA
from .exceptions import *  # NOQA



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test cases for the io handling code.
"""
from __future__ import print_function
from __future__ import division

import tempfile
import unittest

import mock

import htsget


class MockedResponse(object):
    """
    Mocked response object for requests.
    """
    def __init__(self, ticket, data):
        self.ticket = ticket
        self.headers = {}
        self.data = data

    def json(self):
        return self.ticket

    def raise_for_status(self):
        pass

    def iter_content(self, size):
        yield self.data


class MockedRequestsTest(unittest.TestCase):
    """
    Test cases where we mock out requests.get.
    """

    def test_simple_case(self):
        ticket_url = "http://ticket.com"
        data_url = "http://data.url.com"
        headers = {"a": "a", "xyz": "ghj"}
        ticket = {"urls": [
            {"url": data_url, "headers": headers}
        ]}
        data = b"0" * 1024
        returned_response = MockedResponse(ticket, data)
        with mock.patch("requests.get", return_value=returned_response) as mocked_get:
            with tempfile.NamedTemporaryFile("wb+") as f:
                htsget.get(ticket_url, f)
                f.seek(0)
                self.assertEqual(f.read(), data)
            self.assertEqual(mocked_get.call_count, 2)
            # Note that we only get the arguments for the last call using this method.
            args, kwargs = mocked_get.call_args
            self.assertEqual(args[0], data_url)
            self.assertEqual(kwargs["headers"], headers)
            self.assertEqual(kwargs["stream"], True)



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Main interface for the htsget library.
"""
from __future__ import division
from __future__ import print_function

import logging

import htsget.protocol as protocol
import htsget.exceptions as exceptions

import requests

CONTENT_LENGTH = "Content-Length"


def get(
        url, output, reference_name=None, reference_md5=None,
        start=None, end=None, fields=None, tags=None, notags=None,
        data_format=None, max_retries=5, retry_wait=5, timeout=10):
    """
    Runs a request to the specified URL and write the resulting data to
    the specified file-like object.

    :param str url: The URL of the data to retrieve. This may be composed of a prefix
        such as ``http://example.com/reads/`` and an ID suffix such as
        ``NA12878``. The full URL must be supplied here, i.e., in this example
        ``http://example.com/reads/NA12878``.
    :param file output: A file-like object to write the downloaded data to. To support
       retrying of failed transfers, this file must be seekable. For this reason,
       retry will fail if ``stdout`` is provided.
    :param str reference_name: The reference sequence name, for example "chr1",
        "1", or "chrX". If unspecified, all data is returned.
    :param str reference_md5: The MD5 checksum uniquely representing the reference
        sequence as a lower-case hexadecimal string, calculated as the MD5 of the
        upper-case sequence excluding all whitespace characters (this is equivalent to
        SQ:M5 in SAM).
    :param int start: The start position of the range on the reference, 0-based,
        inclusive. If specified, ``reference_name`` or ``reference_md5`` must also
        be specified.
    :param int end: The end position of the range on the reference, 0-based exclusive.
        If specified, ``reference_name`` or ``reference_md5`` must also be specified.
    :param str data_format: The requested format of the returned data.
    :param int max_retries: The maximum number of times that an individual transfer
        will be retried.
    :param float retry_wait: The amount of time in seconds to wait before retrying
        a failed transfer.
    :param float timeout: The socket timeout for I/O operations.
    """
    manager = SynchronousDownloadManager(
        url, output, reference_name=reference_name,
        reference_md5=reference_md5, start=start, end=end, fields=fields, tags=tags,
        notags=notags, data_format=data_format, max_retries=max_retries, timeout=timeout,
        retry_wait=retry_wait)
    manager.run()


class SynchronousDownloadManager(protocol.DownloadManager):
    """
    Class implementing the GA4GH streaming API synchronously using the
    requests library.
    """
    def __get(self, *args, **kwargs):
        try:
            response = requests.get(*args, **kwargs)
        except requests.RequestException as re:
            raise exceptions.RetryableIOError(re)
        try:
            response.raise_for_status()
        except requests.HTTPError as he:
            # TODO classify other errors that we consider unrecoverable.
            if response.status_code == 404:
                raise exceptions.ExceptionWrapper(he)
            else:
                raise exceptions.RetryableIOError(he)
        return response

    def _handle_ticket_request(self):
        logging.debug("handle_ticket_request(url={})".format(self.ticket_request_url))
        response = self.__get(self.ticket_request_url, timeout=self.timeout)
        self.ticket = response.json()

    def _handle_http_url(self, url, headers):
        logging.debug("handle_http_url(url={}, headers={})".format(url, headers))
        response = self.__get(url, headers=headers, stream=True, timeout=self.timeout)
        length = 0
        piece_size = 65536
        try:
            for piece in response.iter_content(piece_size):
                length += len(piece)
                self.output.write(piece)
        except requests.RequestException as re:
            raise exceptions.RetryableIOError(re)
        if CONTENT_LENGTH in response.headers:
            content_length = int(response.headers[CONTENT_LENGTH])
            if content_length != length:
                raise exceptions.ContentLengthMismatch(
                    "Length mismatch {} != {}".format(content_length, length))



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
The command line interface for htsget.
"""

from __future__ import division
from __future__ import print_function


import argparse
import logging
import os
import signal
import sys

import htsget
import htsget.exceptions as exceptions


def error_message(message):
    """
    Writes an error message to stderr.
    """
    print("{}: error: {}".format(sys.argv[0], message), file=sys.stderr)


def run(args):
    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG
    logging.basicConfig(format='%(asctime)s %(message)s', level=log_level)

    if args.output is not None:
        output = open(args.output, "wb")
    else:
        output = sys.stdout
        if args.max_retries != 0:
            logging.warn(
                "Cannot retry failed transfers when writing to stdout. Setting "
                "max_retries to zero")
            args.max_retries = 0
    exit_status = 1
    try:
        htsget.get(
            args.url, output, reference_name=args.reference_name,
            reference_md5=args.reference_md5, start=args.start,
            end=args.end, data_format=args.format, max_retries=args.max_retries,
            retry_wait=args.retry_wait, timeout=args.timeout)
        exit_status = 0
    except exceptions.ExceptionWrapper as ew:
        error_message(str(ew))
    except exceptions.HtsgetException as he:
        error_message(str(he))
    except KeyboardInterrupt:
        error_message("interrupted")
    finally:
        if output is not sys.stdout:
            output.close()
    sys.exit(exit_status)


def get_htsget_parser():
    parser = argparse.ArgumentParser(
        description=(
            "Command line interface for the GA4GH Streaming API. Provides "
            "a simple method to retrieve data spanning genomic ranges from "
            "servers supporting the protocol."))
    parser.add_argument(
        "-V", "--version", action='version',
        version='%(prog)s {}'.format(htsget.__version__))
    parser.add_argument(
        '--verbose', '-v', action='count', default=0,
        help="Increase verbosity.")
    parser.add_argument(
        "url", type=str, help="The URL of the object to retrieve")
    parser.add_argument(
        "--format", "-f", type=str, default=None,
        help="The format of data to request.")
    parser.add_argument(
        "--reference-name", "-r", type=str, default=None,
        help=(
            "The reference sequence name, for example 'chr1', '1', or 'chrX'. "
            "If unspecified, all data is returned."))
    parser.add_argument(
        "--reference-md5", "-m", type=str, default=None,
        help=(
            "The MD5 checksum uniquely representing the requested reference "
            "sequence as a lower-case hexadecimal string, calculated as the MD5 "
            "of the upper-case sequence excluding all whitespace characters."))
    parser.add_argument(
        "--start", "-s", type=int, default=None,
        help=(
            "The start position of the range on the reference, 0-based, inclusive. "
            "If specified, reference-name or reference-md5 must also be specified."))
    parser.add_argument(
        "--end", "-e", type=int, default=None,
        help=(
            "The end position of the range on the reference, 0-based exclusive. If "
            "specified, reference-name or reference-md5 must also be specified."))
    parser.add_argument(
        "--output", "-O", type=str, default=None,
        help=(
            "The output file path. Defaults to stdout. If output is to stdout, the "
            "max-retries is set to zero, since retries are not supported when writing "
            "to this stream"))
    parser.add_argument(
        "--max-retries", "-M", type=int, default=5,
        help="The maximum number of times to retry a failed transfer.")
    parser.add_argument(
        "--retry-wait", "-W", type=float, default=5,
        help="The number of seconds to wait before retrying a failed transfer.")
    parser.add_argument(
        "--timeout", "-T", type=float, default=10,
        help="The socket timeout for transfers.")
    return parser


def htsget_main():
    if os.name == "posix":
        # Set signal handler for SIGPIPE to quietly kill the program.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    parser = get_htsget_parser()
    args = parser.parse_args()
    run(args)



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test cases for the htsget CLI.
"""
from __future__ import print_function
from __future__ import division

import logging
import os
import sys
import tempfile
import unittest

import mock

import htsget.cli as cli
import htsget.exceptions as exceptions


class TestMain(unittest.TestCase):
    """
    Simple tests for the main function.
    """
    with mock.patch("htsget.cli.run") as mocked_run, \
            mock.patch("argparse.ArgumentParser.parse_args") as mocked_parse:
        cli.htsget_main()
        mocked_parse.assert_called_once()
        mocked_run.assert_called_once()


class TestHtsgetArgumentParser(unittest.TestCase):
    """
    Tests the parser to ensure it parses input values correctly.
    """
    def parse_args(self, args):
        parser = cli.get_htsget_parser()
        return parser.parse_args(args)

    def test_defaults(self):
        args = self.parse_args(["URL"])
        self.assertEqual(args.url, "URL")
        self.assertEqual(args.format, None)
        self.assertEqual(args.reference_name, None)
        self.assertEqual(args.reference_md5, None)
        self.assertEqual(args.start, None)
        self.assertEqual(args.end, None)
        self.assertEqual(args.output, None)
        self.assertEqual(args.max_retries, 5)
        self.assertEqual(args.retry_wait, 5)
        self.assertEqual(args.timeout, 10)


class TestHtsgetRun(unittest.TestCase):
    """
    Tests to ensure the run function correctly passes along parameters.
    """

    def setUp(self):
        fd, self.output_filename = tempfile.mkstemp(prefix="htsget_cli_test_")
        os.close(fd)

    def tearDown(self):
        os.unlink(self.output_filename)

    def run_cmd(self, cmd):
        parser = cli.get_htsget_parser()
        args = parser.parse_args(cmd.split())
        with mock.patch("htsget.get") as mocked_get, \
                mock.patch("sys.exit") as mocked_exit:
            cli.run(args)
            self.assertEqual(mocked_get.call_count, 1)
            mocked_exit.assert_called_once_with(0)
            return mocked_get.call_args

    def test_defaults(self):
        url = "http://example.com/stuff"
        args, kwargs = self.run_cmd("{}".format(url))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1], sys.stdout)
        self.assertEqual(kwargs["start"], None)
        self.assertEqual(kwargs["end"], None)
        self.assertEqual(kwargs["reference_name"], None)
        self.assertEqual(kwargs["reference_md5"], None)
        self.assertEqual(kwargs["data_format"], None)

    def test_defaults_with_file(self):
        url = "http://example.com/stuff"
        args, kwargs = self.run_cmd("{} -O {}".format(url, self.output_filename))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1].name, self.output_filename)
        self.assertEqual(kwargs["start"], None)
        self.assertEqual(kwargs["end"], None)
        self.assertEqual(kwargs["reference_name"], None)
        self.assertEqual(kwargs["reference_md5"], None)

    def test_reference_name(self):
        url = "http://example.com/otherstuff"
        for reference_name in ["chr1", "1", "x" * 100]:
            args, kwargs = self.run_cmd("{} -O {} -r {}".format(
                url, self.output_filename, reference_name))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], None)
            self.assertEqual(kwargs["end"], None)
            self.assertEqual(kwargs["reference_name"], reference_name)
            self.assertEqual(kwargs["reference_md5"], None)

            args, kwargs = self.run_cmd("{} -O {} --reference-name {}".format(
                url, self.output_filename, reference_name))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], None)
            self.assertEqual(kwargs["end"], None)
            self.assertEqual(kwargs["reference_name"], reference_name)
            self.assertEqual(kwargs["reference_md5"], None)

    def test_reference_md5(self):
        url = "http://example.com/otherstuff"
        reference_md5 = "d7866be4ab9deb8b26d38a978b0684e3"
        args, kwargs = self.run_cmd("{} -O {} -m {}".format(
            url, self.output_filename, reference_md5))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1].name, self.output_filename)
        self.assertEqual(kwargs["start"], None)
        self.assertEqual(kwargs["end"], None)
        self.assertEqual(kwargs["reference_name"], None)
        self.assertEqual(kwargs["reference_md5"], reference_md5)

        args, kwargs = self.run_cmd("{} -O {} --reference-md5 {}".format(
            url, self.output_filename, reference_md5))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1].name, self.output_filename)
        self.assertEqual(kwargs["start"], None)
        self.assertEqual(kwargs["end"], None)
        self.assertEqual(kwargs["reference_name"], None)
        self.assertEqual(kwargs["reference_md5"], reference_md5)

    def test_start(self):
        url = "http://example.com/otherstuff"
        reference_name = "chr2"
        for start in [0, 100, 2**32]:
            args, kwargs = self.run_cmd("{} -O {} -r {} -s {}".format(
                url, self.output_filename, reference_name, start))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], start)
            self.assertEqual(kwargs["end"], None)
            self.assertEqual(kwargs["reference_name"], reference_name)

            args, kwargs = self.run_cmd("{} -O {} -r {} --start {}".format(
                url, self.output_filename, reference_name, start))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], start)
            self.assertEqual(kwargs["end"], None)
            self.assertEqual(kwargs["reference_name"], reference_name)

    def test_end(self):
        url = "http://example.com/otherstuff"
        reference_name = "chr2"
        for end in [0, 100, 2**32]:
            args, kwargs = self.run_cmd("{} -O {} -r {} -e {}".format(
                url, self.output_filename, reference_name, end))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], None)
            self.assertEqual(kwargs["end"], end)
            self.assertEqual(kwargs["reference_name"], reference_name)

            args, kwargs = self.run_cmd("{} -O {} -r {} --end {}".format(
                url, self.output_filename, reference_name, end))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], None)
            self.assertEqual(kwargs["end"], end)
            self.assertEqual(kwargs["reference_name"], reference_name)

    def test_start_end(self):
        url = "http://example.com/otherstuff"
        reference_name = "chr2"
        for start, end in [(0, 1), (100, 200), (5, 2**32)]:
            args, kwargs = self.run_cmd("{} -O {} -r {} -s {} -e {}".format(
                url, self.output_filename, reference_name, start, end))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["start"], start)
            self.assertEqual(kwargs["end"], end)
            self.assertEqual(kwargs["reference_name"], reference_name)

    def test_format(self):
        url = "http://example.com/otherstuff"
        for fmt in ["bam", "CRAM", "anything"]:
            args, kwargs = self.run_cmd("{} -O {} -f {}".format(
                url, self.output_filename, fmt))
            self.assertEqual(args[0], url)
            self.assertEqual(args[1].name, self.output_filename)
            self.assertEqual(kwargs["data_format"], fmt)

    def test_max_retries(self):
        url = "http://example.com/otherstuff"
        for max_retries in [0, 5, 10]:
            args, kwargs = self.run_cmd("{} -O {} -M {}".format(
                url, self.output_filename, max_retries))
            kwargs["max_retries"] = max_retries

            args, kwargs = self.run_cmd("{} -O {} --max-retries {}".format(
                url, self.output_filename, max_retries))
            kwargs["max_retries"] = max_retries

    def test_retry_wait(self):
        url = "http://example.com/otherstuff"
        for retry_wait in [0, 5, 10, 1.4]:
            args, kwargs = self.run_cmd("{} -O {} -W {}".format(
                url, self.output_filename, retry_wait))
            kwargs["retry_wait"] = retry_wait

            args, kwargs = self.run_cmd("{} -O {} --retry-wait {}".format(
                url, self.output_filename, retry_wait))
            kwargs["retry_wait"] = retry_wait

    def test_timeout(self):
        url = "http://example.com/otherstuff"
        for timeout in [0, 5, 10, 1.4]:
            args, kwargs = self.run_cmd("{} -O {} -r {}".format(
                url, self.output_filename, timeout))
            kwargs["timeout"] = timeout

            args, kwargs = self.run_cmd("{} -O {} --timeout {}".format(
                url, self.output_filename, timeout))
            kwargs["timeout"] = timeout

    def test_stdout_zero_retries(self):
        url = "http://example.com/stuff"
        args, kwargs = self.run_cmd("{}".format(url))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1], sys.stdout)
        self.assertEqual(kwargs["max_retries"], 0)

        # this is true even if we specify it explicitly
        args, kwargs = self.run_cmd("{} --max-retries 10".format(url))
        self.assertEqual(args[0], url)
        self.assertEqual(args[1], sys.stdout)
        self.assertEqual(kwargs["max_retries"], 0)


class TestVerbosity(unittest.TestCase):
    """
    Tests to ensure the verbosity settings work.
    """
    def run_cmd(self, cmd):
        parser = cli.get_htsget_parser()
        args = parser.parse_args(cmd.split())
        with mock.patch("htsget.get") as mocked_get, \
                mock.patch("sys.exit") as mocked_exit, \
                mock.patch("logging.basicConfig") as mocked_log_config:
            cli.run(args)
            self.assertEqual(mocked_get.call_count, 1)
            self.assertEqual(mocked_log_config.call_count, 1)
            mocked_exit.assert_called_once_with(0)
            return mocked_log_config.call_args[1]["level"]

    def test_defaults(self):
        level = self.run_cmd("http://url.com")
        self.assertEqual(level, logging.WARNING)

    def test_repeats(self):
        level = self.run_cmd("http://url.com -v")
        self.assertEqual(level, logging.INFO)
        level = self.run_cmd("http://url.com --verbose")
        self.assertEqual(level, logging.INFO)
        level = self.run_cmd("http://url.com -vv")
        self.assertEqual(level, logging.DEBUG)
        level = self.run_cmd("http://url.com -v -v")
        self.assertEqual(level, logging.DEBUG)
        level = self.run_cmd("http://url.com --verbose --verbose")
        self.assertEqual(level, logging.DEBUG)


class TestRuntimeErrors(unittest.TestCase):
    """
    Test cases to cover the various error conditions that may occur at runtime.
    """
    def assert_exception_writes_error_message(self, exception, message):
        parser = cli.get_htsget_parser()
        args = parser.parse_args(["https://some.url"])
        saved_stderr = sys.stderr
        try:
            with tempfile.TemporaryFile("w+") as tmp_stderr:
                sys.stderr = tmp_stderr
                with mock.patch("htsget.get") as mocked_get, \
                        mock.patch("sys.exit") as mocked_exit, \
                        mock.patch("logging.basicConfig"):
                    mocked_get.side_effect = exception
                    cli.run(args)
                tmp_stderr.seek(0)
                stderr = tmp_stderr.read().strip()
                mocked_exit.assert_called_once_with(1)
        finally:
            sys.stderr = saved_stderr
        self.assertTrue(stderr.endswith(message))

    def test_keyboard_interrupt(self):
        self.assert_exception_writes_error_message(KeyboardInterrupt, "interrupted")

    def test_exception_wrapper(self):
        msg = "some message"
        self.assert_exception_writes_error_message(
            exceptions.ExceptionWrapper(Exception(msg)), msg)

    def test_htsget_exception(self):
        msg = "some other message"
        self.assert_exception_writes_error_message(exceptions.HtsgetException(msg), msg)



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test cases using a simple HTTP server running locally.
"""
from __future__ import print_function
from __future__ import division

import json
import os
import sys
import tempfile
import threading
import unittest

import mock
from six.moves import BaseHTTPServer
from six.moves import socketserver
from six.moves.urllib.parse import urljoin

import htsget
import htsget.exceptions as exceptions
import htsget.cli as cli

PORT = 6160
SERVER_URL = "http://localhost:{}".format(PORT)


class TestUrlInstance(object):
    def __init__(self, url, data, headers={}, error_code=None, truncate=False):
        self.url = url
        self.data = data
        self.headers = headers
        self.error_code = error_code
        self.truncate = truncate


class TestServer(socketserver.TCPServer):
    """
    A local test server designed to be run in a thread and shutdown smoothly
    for test cases.
    """
    allow_reuse_address = True
    # This is set by clients to contain the list of TestUrlInstance objects.
    test_instances = []

    def shutdown(self):
        self.socket.close()
        socketserver.TCPServer.shutdown(self)


class TestRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    ticket_path = "/ticket"
    ticket_url = urljoin(SERVER_URL, ticket_path)

    def log_message(self, format, *args):
        # Silence the logger.
        pass

    def do_GET(self):
        url_map = {instance.url: instance for instance in self.server.test_instances}
        if self.path == self.ticket_path:
            self.send_response(200)
            self.end_headers()
            urls = [
                {
                    "url": urljoin(SERVER_URL, test_instance.url),
                    "headers": test_instance.headers
                } for test_instance in self.server.test_instances]
            ticket = {
                "urls": urls
            }
            self.wfile.write(json.dumps(ticket).encode())
        elif self.path in url_map:
            instance = url_map[self.path]
            if instance.error_code is not None:
                self.send_error(instance.error_code)
            else:
                self.send_response(200)
                self.send_header("Content-Length", len(instance.data))
                if instance.truncate:
                    self.end_headers()
                    self.wfile.write(instance.data[:-1])
                else:
                    self.end_headers()
                    self.wfile.write(instance.data)
        else:
            self.send_error(404)


class ServerTest(unittest.TestCase):
    """
    Superclass of tests needing a server running in a thread.
    """
    @classmethod
    def setup_class(cls):
        cls.httpd = TestServer(("", PORT), TestRequestHandler)

        def target():
            # Sometimes the server doesn't shutdown cleanly, but we don't
            # care here.
            try:
                cls.httpd.serve_forever()
            except ValueError:
                pass

        cls.httpd_thread = threading.Thread(target=target)
        cls.httpd_thread.start()

    @classmethod
    def teardown_class(cls):
        cls.httpd.shutdown()
        cls.httpd_thread.join()

    def setUp(self):
        self.output_file = tempfile.NamedTemporaryFile("wb+")


class TestDataTransfers(ServerTest):
    """
    Test cases for various data transfers.
    """
    def assert_data_transfer_ok(self, test_instances, max_retries=0):
        self.httpd.test_instances = test_instances
        htsget.get(
            TestRequestHandler.ticket_url, self.output_file, max_retries=max_retries)
        self.output_file.seek(0)
        all_data = b"".join(test_instance.data for test_instance in test_instances)
        self.assertEqual(self.output_file.read(), all_data)

    def test_simple_data(self):
        instances = [
            TestUrlInstance(url="/data1", data=b"data1"),
            TestUrlInstance(url="/data2", data=b"data2")
        ]
        self.assert_data_transfer_ok(instances)

    def test_binary_data(self):
        instances = []
        for j in range(10):
            instances.append(TestUrlInstance(
                url="/path/to/data/{}".format(j),
                data=bytes(j) * 1024))
        self.assert_data_transfer_ok(instances)

    def test_transfer_with_cli(self):
        test_instances = [
            TestUrlInstance(url="/data1", data=b"data1"),
            TestUrlInstance(url="/data2", data=b"data2")
        ]
        self.httpd.test_instances = test_instances
        try:
            fd, filename = tempfile.mkstemp()
            os.close(fd)
            cmd = [TestRequestHandler.ticket_url, "-O", filename]
            parser = cli.get_htsget_parser()
            args = parser.parse_args(cmd)
            with mock.patch("sys.exit") as mocked_exit:
                cli.run(args)
                mocked_exit.assert_called_once_with(0)
            all_data = b"".join(test_instance.data for test_instance in test_instances)
            with open(filename, "rb") as f:
                self.assertEqual(f.read(), all_data)
        finally:
            os.unlink(filename)

    def test_transfer_with_cli_stdout(self):
        test_instances = [
            TestUrlInstance(url="/data1", data=b"data1"),
            TestUrlInstance(url="/data2", data=b"data2")
        ]
        self.httpd.test_instances = test_instances
        saved = sys.stdout
        sys.stdout = self.output_file
        try:
            cmd = [TestRequestHandler.ticket_url]
            parser = cli.get_htsget_parser()
            args = parser.parse_args(cmd)
            with mock.patch("sys.exit") as mocked_exit:
                cli.run(args)
                mocked_exit.assert_called_once_with(0)
            all_data = b"".join(test_instance.data for test_instance in test_instances)
            self.output_file.seek(0)
            self.assertEqual(self.output_file.read(), all_data)
        finally:
            sys.stdout = saved


class TestErrorHandling(ServerTest):
    """
    Test cases for the error handling in the client.
    """
    def test_missing_path(self):
        self.assertRaises(
            exceptions.ExceptionWrapper, htsget.get, SERVER_URL + "/nopath",
            self.output_file, max_retries=0)

    def test_bad_port(self):
        self.assertRaises(
            exceptions.RetryableIOError, htsget.get, "http://localhost:66123",
            self.output_file, max_retries=0)

    def test_data_error(self):
        self.httpd.test_instances = [
            TestUrlInstance(url="/fail1", data=b"", error_code=500)
        ]
        self.assertRaises(
            exceptions.RetryableIOError, htsget.get,
            TestRequestHandler.ticket_url, self.output_file, max_retries=0)

    def test_data_truncation(self):
        self.httpd.test_instances = [
            TestUrlInstance(url="/fail1", data=b"x" * 8192, truncate=True)
        ]
        self.assertRaises(
            exceptions.ContentLengthMismatch, htsget.get,
            TestRequestHandler.ticket_url, self.output_file, max_retries=0)



from __future__ import division
from __future__ import print_function

from setuptools import setup

with open("README.txt") as f:
    long_description = f.read()

setup(
    name="htsget",
    description="Python API and command line interface for the GA4GH streaming API.",
    long_description=long_description,
    packages=["htsget"],
    author="Jerome Kelleher",
    author_email="jerome.kelleher@well.ox.ac.uk",
    url="http://pypi.python.org/pypi/htsget",
    entry_points={
        'console_scripts': [
            'htsget=htsget.cli:htsget_main',
        ]
    },
    install_requires=["requests", "six"],
    keywords=["BAM", "CRAM", "Streaming"],
    license="Apache Software License",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    setup_requires=['setuptools_scm'],
    use_scm_version={"write_to": "htsget/_version.py"},
)



#
# Copyright 2016 University of Oxford
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test cases for the protocol handling code.
"""
from __future__ import print_function
from __future__ import division

import base64
import collections
import tempfile
import unittest

import mock

from six import StringIO
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import parse_qs

import htsget.protocol as protocol
import htsget.exceptions as exceptions

EXAMPLE_URL = "http://example.com"


class TestManagerAbstractMethods(unittest.TestCase):
    """
    Tests to make sure the abstract methods are correct interfaces.
    """
    def test_not_implemented(self):
        dm = protocol.DownloadManager(EXAMPLE_URL, None)
        self.assertRaises(NotImplementedError, dm._ticket_request)
        self.assertRaises(NotImplementedError, dm._handle_http_url, {})


class TestTicketRequestUrls(unittest.TestCase):
    """
    Tests the ticket request generator.
    """
    def test_defaults(self):
        url = protocol.ticket_request_url(EXAMPLE_URL)
        self.assertEqual(url, EXAMPLE_URL)

    def test_reference_name(self):
        full_url = protocol.ticket_request_url(
            "http://example.co.uk/path/to/resource", reference_name="1",
            start=2, end=100)
        parsed = urlparse(full_url)
        self.assertEqual(parsed.scheme, "http")
        self.assertEqual(parsed.netloc, "example.co.uk")
        self.assertEqual(parsed.path, "/path/to/resource")
        query = parse_qs(parsed.query)
        self.assertEqual(query["referenceName"], ["1"])
        self.assertEqual(query["start"], ["2"])
        self.assertEqual(query["end"], ["100"])
        self.assertEqual(len(query), 3)

    def test_reference_md5(self):
        md5 = "b9185d4fade27aa27e17f25fafec695f"
        full_url = protocol.ticket_request_url(
            "https://example.com/resource", reference_md5=md5)
        parsed = urlparse(full_url)
        self.assertEqual(parsed.scheme, "https")
        self.assertEqual(parsed.netloc, "example.com")
        self.assertEqual(parsed.path, "/resource")
        query = parse_qs(parsed.query)
        self.assertEqual(query["referenceMd5"], [md5])
        self.assertEqual(len(query), 1)

    def test_format(self):
        for data_format in ["cram", "CRAM", "BAM"]:
            full_url = protocol.ticket_request_url(
                "http://example.co.uk/path/to/resource", data_format=data_format)
            parsed = urlparse(full_url)
            self.assertEqual(parsed.scheme, "http")
            self.assertEqual(parsed.netloc, "example.co.uk")
            self.assertEqual(parsed.path, "/path/to/resource")
            query = parse_qs(parsed.query)
            self.assertEqual(query["format"], [data_format.upper()])

    def test_url_scheme(self):
        full_url = protocol.ticket_request_url("http://a.com")
        self.assertEqual(urlparse(full_url).scheme, "http")
        full_url = protocol.ticket_request_url("https://a.com")
        self.assertEqual(urlparse(full_url).scheme, "https")

    def test_url_netloc(self):
        full_url = protocol.ticket_request_url("http://a.com")
        self.assertEqual(urlparse(full_url).netloc, "a.com")
        full_url = protocol.ticket_request_url("http://a.com/other/stuff")
        self.assertEqual(urlparse(full_url).netloc, "a.com")
        full_url = protocol.ticket_request_url("https://192.168.0.1")
        self.assertEqual(urlparse(full_url).netloc, "192.168.0.1")
        full_url = protocol.ticket_request_url("https://192.168.0.1:8080/xyz")
        self.assertEqual(urlparse(full_url).netloc, "192.168.0.1:8080")

    def test_url_path(self):
        full_url = protocol.ticket_request_url("http://a.com/path/to/resource")
        self.assertEqual(urlparse(full_url).path, "/path/to/resource")
        full_url = protocol.ticket_request_url("http://a.com/")
        self.assertEqual(urlparse(full_url).path, "/")
        full_url = protocol.ticket_request_url("http://a.com")
        self.assertEqual(urlparse(full_url).path, "")

    def test_embedded_query_strings(self):
        full_url = protocol.ticket_request_url("http://a.com/stuff?a=a&b=b")
        query = parse_qs(urlparse(full_url).query)
        self.assertEqual(query["a"], ["a"])
        self.assertEqual(query["b"], ["b"])

        full_url = protocol.ticket_request_url(
            "http://a.com/stuff?a=a&b=b", reference_name="123")
        query = parse_qs(urlparse(full_url).query)
        self.assertEqual(query["a"], ["a"])
        self.assertEqual(query["b"], ["b"])
        self.assertEqual(query["referenceName"], ["123"])


def get_http_ticket(url, headers={}):
    return {"url": url, "headers": headers}


def get_data_uri_ticket(url):
    return {"url": url}


def get_ticket(urls=[], format_=None, md5=None):
    d = {"urls": urls}
    if format_ is not None:
        d["format"] = format_
    if md5 is not None:
        d["md5"] = md5
    return d


class TestDownloadManager(protocol.DownloadManager):
    """
    Test download manager that stores a ticket.
    """
    def __init__(self, test_ticket, output, **kwargs):
        super(TestDownloadManager, self).__init__(EXAMPLE_URL, output, **kwargs)
        self.test_ticket = test_ticket

    def _handle_ticket_request(self):
        self.ticket = self.test_ticket


class StoringUrlsDownloadManager(TestDownloadManager):
    """
    Simple implementation of the DownloadManager that just saves the URLs.
    """
    def __init__(self, test_ticket, output):
        super(StoringUrlsDownloadManager, self).__init__(test_ticket, output)
        self.stored_urls = []

    def _handle_data_uri(self, parsed_url):
        self.stored_urls.append(parsed_url)

    def _handle_http_url(self, url, headers):
        self.stored_urls.append((url, headers))


class TestTicketResponses(unittest.TestCase):
    """
    Test cases for dealing with the responses from a given server.
    """
    def test_bad_scheme(self):
        with tempfile.TemporaryFile("w+") as temp_file:
            for bad_scheme in ["htt://as", "file:///home", "ftp://x.y/sdf"]:
                ticket = get_ticket(urls=[
                    get_http_ticket("http://a.b"),
                    get_http_ticket("htp")])
                dm = StoringUrlsDownloadManager(ticket, temp_file)
                self.assertRaises(ValueError, dm.run)

    def test_basic_http_parsing(self):
        headers = {"a": "b", "b": "c"}
        ticket = get_ticket(urls=[get_http_ticket(EXAMPLE_URL, headers)])
        with tempfile.TemporaryFile("w+") as temp_file:
            dm = StoringUrlsDownloadManager(ticket, temp_file)
            dm.run()
        self.assertEqual(dm.stored_urls[0], (EXAMPLE_URL, headers))

    def test_basic_data_uri_parsing(self):
        data_uri = "data:application/vnd.ga4gh.bam;base64,SGVsbG8sIFdvcmxkIQ=="
        ticket = get_ticket(urls=[get_data_uri_ticket(data_uri)])
        with tempfile.TemporaryFile("w+") as temp_file:
            dm = StoringUrlsDownloadManager(ticket, temp_file)
            dm.run()
        self.assertEqual(dm.stored_urls[0], urlparse(data_uri))


class TestDataUriParsing(unittest.TestCase):
    """
    Tests for the data URI scheme
    """
    def test_basic_parsing(self):
        data_uri = "data:application/vnd.ga4gh.bam;base64,SGVsbG8sIFdvcmxkIQ=="
        ticket = get_ticket(urls=[get_data_uri_ticket(data_uri)])
        with tempfile.TemporaryFile("wb+") as temp_file:
            dm = TestDownloadManager(ticket, temp_file)
            dm.run()
            temp_file.seek(0)
            data = temp_file.read()
            other_data = base64.b64decode("SGVsbG8sIFdvcmxkIQ==")
            self.assertEqual(data, other_data)

    # TODO much more tests here with input data. Also need to test the various
    # chunks of the parser.


class RetryCountDownloadManager(TestDownloadManager):

    def __init__(self, test_ticket, output, **kwargs):
        super(RetryCountDownloadManager, self).__init__(test_ticket, output, **kwargs)
        self.attempt_counts = collections.Counter()

    def _handle_http_url(self, url, headers):
        self.attempt_counts[url] += 1
        raise exceptions.RetryableError()


class FailingChunkDownloadManager(TestDownloadManager):

    def __init__(self, test_ticket, output, data_map, **kwargs):
        super(FailingChunkDownloadManager, self).__init__(test_ticket, output, **kwargs)
        self.attempt_counts = collections.Counter()
        self.data_map = data_map

    def _handle_http_url(self, url, headers):
        self.attempt_counts[url] += 1
        if self.attempt_counts[url] == 1:
            self.output.write("gibberish" * 100)
            raise exceptions.RetryableError()
        else:
            self.output.write(self.data_map[url])


class TestRetries(unittest.TestCase):
    """
    Test the retry logic.
    """
    def test_num_retries(self):
        ticket = get_ticket(urls=[get_http_ticket(EXAMPLE_URL)])
        with tempfile.TemporaryFile("w+") as temp_file:
            for num_retries in range(10):
                with mock.patch("time.sleep") as mock_sleep, \
                        mock.patch("logging.warning") as mock_warning:
                    dm = RetryCountDownloadManager(
                        ticket, temp_file, max_retries=num_retries)
                    self.assertEqual(dm.max_retries, num_retries)
                    self.assertRaises(exceptions.RetryableError, dm.run)
                    self.assertEqual(dm.attempt_counts[EXAMPLE_URL], num_retries + 1)
                    self.assertEqual(mock_sleep.call_count, num_retries)
                    self.assertEqual(mock_warning.call_count, num_retries)

    def test_retry_wait(self):
        ticket = get_ticket(urls=[get_http_ticket(EXAMPLE_URL)])
        with tempfile.TemporaryFile("w+") as temp_file:
            for sleep_time in range(10):
                with mock.patch("time.sleep") as mock_sleep:
                    dm = RetryCountDownloadManager(
                        ticket, temp_file, max_retries=1, retry_wait=sleep_time)
                    self.assertEqual(dm.retry_wait, sleep_time)
                    self.assertEqual(dm.max_retries, 1)
                    self.assertRaises(exceptions.RetryableError, dm.run)
                    mock_sleep.assert_called_once_with(sleep_time)

    def test_retry_data(self):
        data_map = {
            "http://url1.com": "1" * 1000,
            "http://url2.com": "2" * 1000,
            "http://url3.com": "3" * 1000
        }
        real_data = "".join(data_map.values())
        ticket = get_ticket(urls=[get_http_ticket(url) for url in data_map.keys()])
        with tempfile.TemporaryFile("w+") as temp_file:
            with mock.patch("time.sleep") as mock_sleep:
                dm = FailingChunkDownloadManager(ticket, temp_file, data_map)
                dm.run()
                temp_file.seek(0)
                self.assertEqual(real_data, temp_file.read())
                self.assertEqual(mock_sleep.call_count, len(data_map) * 1)

    def test_unseekable_file(self):
        def tell_fails():
            raise IOError()
        ticket = get_ticket(urls=[get_http_ticket(EXAMPLE_URL)])
        for num_retries in range(10):
            temp_file = StringIO()
            temp_file.tell = tell_fails
            with mock.patch("time.sleep") as mock_sleep, \
                    mock.patch("logging.warning") as mock_warning:
                dm = RetryCountDownloadManager(
                    ticket, temp_file, max_retries=num_retries)
                self.assertEqual(dm.max_retries, num_retries)
                self.assertRaises(exceptions.RetryableError, dm.run)
                self.assertEqual(dm.attempt_counts[EXAMPLE_URL], 1)
                self.assertEqual(mock_sleep.call_count, 0)
                self.assertEqual(mock_warning.call_count, 0)
